/*
*  PgpSdk1.cpp
*  Copyright 1999-2002 by Benchmark Software
*  All Rights Reserved
*
*  This source file is released for peer review purposes only. It is not public domain.
*  Redistribution either in part or whole is forbidden. The sole source for QuickSilver
*  distribution currently is http://quicksilver.skuz.net
*
*  Richard
*  --
*  R.Christman
*  Benchmark Software
*  quicksilver@skuz.net
*  http://quicksilver.skuz.net
*
*/
#define PGP_WIN32
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
#include <math.h>

#include "PgpSdk6.h"
#include "PGPfiles.h"
#include "PgpMsg.h"
#include "wipedlgs.h"

TPGPsdk *PGPsdk=NULL;
HINSTANCE hPgpSdk=NULL;
TForm *MsgBoxOwner=NULL;
struct Tdecinfo {
   TStrings *siginfos;
   int decrypts;
   int sigs;
} decinfo;

/*
* SDK API version
* example: 1.0.0 = 0x01000000
*   0x01000000   (16777216)   SDK 1.0.0   PGP550
*   0x02000000   (33554432)   SDK 1.1.0
*   0x02000010   (33554448)   SDK 1.1.1   PGP553
*   0x03000000   (50331648)   SDK 1.5.0   PGP600
*   0x03000010   (50331664)   SDK 1.5.1
*   0x03000020   (50331680)   SDK 1.5.2   PGP602
*   0x03001000   (50335744)   SDK 1.6.0
*   0x03002000   (50339840)   SDK 1.7.0   PGP650
*   0x03002010   (50339856)   SDK 1.7.1   PGP651,2,3   // changed pref selectors
*   0x--------   (--------)   SDK 1.7.8   PGP658
*/
//---------------------------------------------------------------------------
bool __fastcall HasSdk()
{
   return hPgpSdk!=NULL;
}
//---------------------------------------------------------------------------
int  __fastcall SdkVer()
{
   if (PGPsdk) return PGPsdk->sdkversion;
   return 0;
}
//---------------------------------------------------------------------------
int  __fastcall SdkApiVer()
{
   if (PGPsdk) return PGPsdk->apiversion;
   return 0;
}
//---------------------------------------------------------------------------
// TPGPsdk
//---------------------------------------------------------------------------

__fastcall TPGPsdk::TPGPsdk()
   : TObject()
{
}
//---------------------------------------------------------------------------

__fastcall TPGPsdk::TPGPsdk(TForm *Owner, HINSTANCE &hPgpSdk)
   : TObject()
{
   owner = Owner;
   hSdkLib = hPgpSdk;
   MsgBoxOwner = Owner;
   headers = new TStringList();
   siginfos = new TStringList();
   pubkeys = new TPGPkeylist(PgpPublicRing());
   SetCurrentDirectory(curdir);
   InitFailed = true;

   err=PGPsdkInit();
   if (IsPgpError(err,"Unable to initialize PGPsdk!"))
      return;

   err=PGPGetSDKVersion(&sdkversion);
   if (IsPgpError(err,"Failed to establish SDK version!"))
      return;

   err=PGPNewContext(kPGPsdkAPIVersion,&context);
   if (IsPgpError(err,"Failed to initialize PGP context structure!"))
      return; // false;

   err=PGPsdkLoadDefaultPrefs(context);
   if (IsPgpError(err,"Unable to load PGPsdk default preferences!"))
      return; // false;

   PGPFileSpecRef pubSpec = kInvalidPGPFileSpecRef;
   PGPFileSpecRef secSpec = kInvalidPGPFileSpecRef;

   // if first attempt fails, use prefSelector from earlier sdk version
   err=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)5,&pubSpec);
   if (err==-12000) err=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)0,&pubSpec);
   if (IsPgpError(err,"Unable to get PGP public keyring file spec!"))
      goto exit;

   err=PGPGetFullPathFromFileSpec(pubSpec,&pubringpkr);
   if (IsPgpError(err,"Unable to get PGP public keyring path!"))
      goto exit;

   // if first attempt fails, use prefSelector from earlier sdk version
   err=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)6,&secSpec);
   if (err==-12000) err=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)1,&secSpec);
   if (IsPgpError(err,"Unable to get PGP secret keyring file spec!"))
      goto exit;

   err=PGPGetFullPathFromFileSpec(secSpec,&secringskr);
   if (IsPgpError(err,"Unable to get PGP secret keyring path!"))
      goto exit;

   err=PGPGetSDKVersion(&apiversion);
   if (IsPgpError(err,"Failed to get PGPsdk API version!"))
      goto exit;

   StrLower(pubringpkr);
   StrLower(secringskr);
   InitFailed = false;

exit:
   if (PGPFileSpecRefIsValid(secSpec)) PGPFreeFileSpec(secSpec);
   if (PGPFileSpecRefIsValid(pubSpec)) PGPFreeFileSpec(pubSpec);

   TPGPkeylist *remkeys = new TPGPkeylist(PgpRemailRing());
   pubkeys->AddKeys(remkeys);
   delete remkeys;
}
//---------------------------------------------------------------------------
__fastcall TPGPsdk::~TPGPsdk()
{
   delete pubkeys;
   delete siginfos;
   delete headers;

   if (!InitFailed)
   {
      err = PGPsdkCleanup();
      IsPgpError(err,"Unable to cleanup PGPsdk for exit!");
   }
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::ExportDefaultKeys(TStrings *exports)
{
   SetCurrentDirectory(curdir);
   exports->Clear();
   TStringList *revokes = new TStringList();
   TStringList *disables = new TStringList();
   PGPKeySetRef keyset = kInvalidPGPKeySetRef;

   err=PGPOpenDefaultKeyRings(context,false,&keyset);
   if (IsPgpError(err,"Failed to open default keyrings!"))
      goto exit;

   err=GetRevokedKeyIds(keyset,revokes);
   if (IsPGPError(err))
      goto exit;

   err=GetDisableKeyIds(keyset,disables);
   if (IsPGPError(err))
      goto exit;

   err=ExportKeySet(keyset,kPGPExportFormat_Basic,true,exports); // true = include privates
   if (IsPGPError(err))
      goto exit;

   if (!disables->Text.IsEmpty()) exports->Insert(0,"Disabled: "+disables->CommaText);
   if (!revokes->Text.IsEmpty())exports->Insert(0,"Revoked: "+revokes->CommaText);

exit:
   if (PGPKeySetRefIsValid(keyset)) PGPFreeKeySet(keyset);
   delete disables;
   delete revokes;
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::ExportDefaultKeysFast(TStrings *exports)
{
   exports->Clear();
   PGPKeySetRef keyset = kInvalidPGPKeySetRef;
   PGPKeySetRef nodisables = kInvalidPGPKeySetRef;
   PGPFilterRef disfilter = kInvalidPGPFilterRef;
   PGPKeySetRef norevokes = kInvalidPGPKeySetRef;
   PGPFilterRef revfilter = kInvalidPGPFilterRef;

   err=PGPOpenDefaultKeyRings(context,false,&keyset);
   if (IsPgpError(err,"Failed to open default keyrings!"))
      goto exit;

   err = PGPNewKeyRevokedFilter(context,false,&revfilter);
   if (IsPgpError(err,"Failed to create revoked key filter."))
      goto exit;

   err = PGPFilterKeySet(keyset,revfilter,&norevokes);
   if (IsPgpError(err,"Failed to get keys from revoked filter."))
      goto exit;

   err = PGPNewKeyDisabledFilter(context,false,&disfilter);
   if (IsPgpError(err,"Failed to create disabled key filter."))
      goto exit;

   err = PGPFilterKeySet(norevokes,disfilter,&nodisables);
   if (IsPgpError(err,"Failed to get keys from disabled filter."))
      goto exit;

   err=ExportKeySet(nodisables,kPGPExportFormat_Basic,true,exports);
   if (IsPGPError(err))                           // true = include privates
      goto exit;

exit:
   if (PGPFilterRefIsValid(revfilter)) PGPFreeFilter(revfilter);
   if (PGPKeySetRefIsValid(norevokes)) PGPFreeKeySet(norevokes);
   if (PGPFilterRefIsValid(disfilter)) PGPFreeFilter(disfilter);
   if (PGPKeySetRefIsValid(nodisables)) PGPFreeKeySet(nodisables);
   if (PGPKeySetRefIsValid(keyset)) PGPFreeKeySet(keyset);
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::ExportKeySet(PGPKeySetRef keys, PGPExportFormat &ExportFormat,
   bool IncludePrivates, TStrings *exports)
{
   SetCurrentDirectory(curdir);
   exports->Clear();
   void *data=NULL; PGPSize size;

   err=PGPExportKeySet(keys,
      PGPOAllocatedOutputBuffer(context,&data,-1,&size), // -1 = no size constraint
      PGPOExportFormat(context,ExportFormat),
      PGPOExportPrivateKeys(context,IncludePrivates),
      PGPOVersionString(context,"N/A"),
      PGPOLastOption(context)
   );
   if (IsPgpError(err,"Failed to export keys!"))
      goto exit;

   exports->SetText((char*)data);
exit:
   if (data) PGPFreeData(data);
   return err;
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::GetDisableKeyIds(PGPKeySetRef keyset, TStrings *keyids)
{
   SetCurrentDirectory(curdir);
   keyids->Clear();
   PGPKeyRef key = kInvalidPGPKeyRef;
   PGPFilterRef filter = kInvalidPGPFilterRef;
   PGPKeySetRef diskeys = kInvalidPGPKeySetRef;
   PGPKeyListRef diskeyslist = kInvalidPGPKeyListRef;
   PGPKeyIterRef listIterator = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyid;

   err = PGPNewKeyDisabledFilter(context,true,&filter);
   if (IsPgpError(err,"Failed to create disabled key filter."))
      goto exit;

   err = PGPFilterKeySet(keyset,filter,&diskeys);
   if (IsPgpError(err,"Failed to get disabled keys."))
      goto exit;

   err = PGPOrderKeySet(diskeys,kPGPAnyOrdering,&diskeyslist);
   if (IsPgpError(err,"Failed to create disabled key list."))
      goto exit;

   err=PGPNewKeyIter(diskeyslist,&listIterator);
   if (IsPgpError(err,"Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err=PGPKeyIterNext(listIterator,&key)) && key)
   {
      err=PGPGetKeyIDFromKey(key,&keyid);
      if (IsPgpError(err,"Failed to get key id from revoked key."))
         goto exit;

      err=PGPGetKeyIDString(&keyid,kPGPKeyIDString_Abbreviated,keyidstr);
      if (IsPgpError(err,"Failed to get key id string!"))
         goto exit;

      keyids->Add(keyidstr);
   }
   if (err==kPGPError_EndOfIteration) err = kPGPError_NoErr;
   if (IsPgpError(err,"Failed to complete iteration of revoked key list."))
      goto exit;

exit:
   if (PGPKeyIterRefIsValid(listIterator)) PGPFreeKeyIter(listIterator);
   if (PGPKeyListRefIsValid(diskeyslist)) PGPFreeKeyList(diskeyslist);
   if (PGPKeySetRefIsValid(diskeys)) PGPFreeKeySet(diskeys);
   if (PGPFilterRefIsValid(filter)) PGPFreeFilter(filter);
   return err;
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::GetRevokedKeyIds(PGPKeySetRef keyset, TStrings *keyids)
{
   SetCurrentDirectory(curdir);
   keyids->Clear();
   PGPKeyRef key = kInvalidPGPKeyRef;
   PGPFilterRef filter = kInvalidPGPFilterRef;
   PGPKeySetRef revkeys = kInvalidPGPKeySetRef;
   PGPKeyListRef revkeyslist = kInvalidPGPKeyListRef;
   PGPKeyIterRef listIterator = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyid;

   err = PGPNewKeyRevokedFilter(context,true,&filter);
   if (IsPgpError(err,"Failed to create revoked key filter."))
      goto exit;

   err = PGPFilterKeySet(keyset,filter,&revkeys);
   if (IsPgpError(err,"Failed to get revoked keys."))
      goto exit;

   err = PGPOrderKeySet(revkeys,kPGPAnyOrdering,&revkeyslist);
   if (IsPgpError(err,"Failed to create revoked key list."))
      goto exit;

   err=PGPNewKeyIter(revkeyslist, &listIterator);
   if (IsPgpError(err,"Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err=PGPKeyIterNext(listIterator,&key)) && key)
   {
      err=PGPGetKeyIDFromKey(key,&keyid);
      if (IsPgpError(err,"Failed to get key id from revoked key."))
         goto exit;

      err=PGPGetKeyIDString(&keyid,kPGPKeyIDString_Abbreviated,keyidstr);
      if (IsPgpError(err,"Failed to get key id string!"))
         goto exit;

      keyids->Add(keyidstr);
   }
   if (err==kPGPError_EndOfIteration) err = kPGPError_NoErr;
   if (IsPgpError(err,"Failed to complete iteration of revoked key list."))
      goto exit;

exit:
   if (PGPKeyIterRefIsValid(listIterator)) PGPFreeKeyIter(listIterator);
   if (PGPKeyListRefIsValid(revkeyslist)) PGPFreeKeyList(revkeyslist);
   if (PGPKeySetRefIsValid(revkeys)) PGPFreeKeySet(revkeys);
   if (PGPFilterRefIsValid(filter)) PGPFreeFilter(filter);
   return err;
}
//---------------------------------------------------------------------------
void __fastcall TPGPsdk::ShowPGPError(PGPError err, AnsiString errmsg)
{
   if(IsPGPError(err))
   {
  	  char *buf = new char[256];
 	  PGPGetErrorString(err, 255, buf);
      MessageBox(owner->Handle,
         AnsiString(errmsg+"\nError: ("+AnsiString(err)+") "+AnsiString(buf)).c_str(),
        "PGP Error",
         MB_OK
      );
      delete buf;
   }
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::CheckPGP6NeedsUpdate(TControl *ForHourGlass)
{
   if (!PGP6AUTOUPDATE) return false;

   int QDKpub = FileAge(pubringpkr); // QDK = sdk keyrings
   int PGPpub = FileAge(PgpPublicRing());    // quicksilver keyrings
   int QDKsec = FileAge(secringskr);
   int PGPsec = FileAge(PgpSecretRing());
   if (QDKpub<0 || QDKsec<0) return false;   // -1 returned if file does not exist
   if (PGPpub<0 || PGPsec<0 || QDKpub>PGPpub || QDKsec>PGPsec)
   {
      UpdateQSKeyrings(ForHourGlass);
      return true;
   }
   return false;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::UpdateQSKeyrings(TControl *ForHourGlass)
{
   SetCurrentDirectory(curdir);
   if (USEPGPKEYSEXE) return UpdateQSKeyringsFast(ForHourGlass);
   else return UpdateQSKeyringsSlow(ForHourGlass);
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::UpdateQSKeyringsSlow(TControl *ForHourGlass)
{
   owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Updating...");
   if (ForHourGlass) ForHourGlass->Cursor = crHourGlass;
   TStringList *block = new TStringList();
   ExportDefaultKeys(block);
   if (!block->Count)
   {
      if (ForHourGlass) ForHourGlass->Cursor = crDefault;
      MsgBoxHwnd(owner->Handle,"Failed to import keys!","File Error",MB_OK);
      return false;
   }

   AnsiString line,rids,dids,id;
   while (block->Count)
   {
      line = block->Strings[0];
      if (line.Pos("-----BEGIN PGP")==1) break;
      if (line.Pos("Disabled: ")==1) dids = line.SubString(11,1024);
      if (line.Pos("Revoked: " )==1) rids = line.SubString(10,1024);
      block->Delete(0);
   }
   if (block->Strings[0].Pos("-----BEGIN PGP")==1)
   {
      TPGPbuffer *secrets,*publics,*newkeys,*pubs,*secs,*key,*chk,*pkt,*pub;

      pkt = new TPGPbuffer();
      pub = new TPGPbuffer();
      key = new TPGPbuffer();
      chk = new TPGPbuffer();
      secrets = new TPGPbuffer();
      publics = new TPGPbuffer();
      newkeys = new TPGPbuffer(block);
      // dispurse to publics and secrets
      while (key->copykeyfrom(newkeys,REWIND))
      {
         if (key->ispublic)
         {
            // add undefined trust packets back to keys
            while (pkt->copypacketfrom(key,REWIND))
            {
               pub->copyfrom(pkt);
               byte keytrust[3] = { 176,1,0 };
               byte sigtrust[3] = { 176,1,0 };
               byte uidtrust[3] = { 176,1,0 };
               if (PublicKeyPacket(pkt->CTB)) pub->Write(keytrust,3); else
               if (SignaturePacket(pkt->CTB)) pub->Write(sigtrust,3); else
               if (UserIdPacket(pkt->CTB)) pub->Write(uidtrust,3);
               pkt->Clear();
            }
            if (dids.Pos("0x"+pub->idtext8)>=1) pub->disable();
            if (rids.Pos("0x"+pub->idtext8)<1)
                publics->copyfrom(pub);
            pub->clear();
         }
         else
            secrets->copyfrom(key);
         key->clear();
      }
      // load publics and add to new keys if not already present
      TStringList *uids = new TStringList();
      uids->Duplicates = dupIgnore;
      publics->getuserids(uids);
      pubs = new TPGPbuffer(PgpPublicRing());
      while (key->copykeyfrom(pubs,REWIND))
      {  // this actions ends with unmatched current keys copied to publics which
         if (uids->IndexOf(key->idtext8)>=0) // becomes the new keyring
            publics->copyfrom(key);
         key->clear();
      }
      // load secrets and add to new keys if not already present
      uids->Text = secrets->getuserids();
      secs = new TPGPbuffer(PgpSecretRing());
      while (key->copykeyfrom(secs,REWIND))
      {
         if (uids->IndexOf(key->idtext8)>=0)
            secrets->copyfrom(key);
         key->clear();
      }
      // save new keyrings to file
      AnsiString pubring = PgpPublicRing();
      AnsiString secring = PgpSecretRing();
      if (pubring.IsEmpty()) pubring = AppDir+"pubring.pkr";
      if (secring.IsEmpty()) secring = AppDir+"secring.skr";
      publics->SaveToFile(pubring);
      secrets->SaveToFile(secring);

      delete secs;
      delete pubs;
      delete key;
      delete chk;
      delete pub;
      delete pkt;
   }
   delete block;
   //reverse earlier visual cues
   if (ForHourGlass) ForHourGlass->Cursor = crDefault;
   //if (Owner) Owner->Caption = cap;
   owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"");
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::UpdateQSKeyringsFast(TControl *ForHourGlass)
{
   owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Updating...");
   if (ForHourGlass) ForHourGlass->Cursor = crHourGlass;

   TStringList *blocks = new TStringList();
   ExportDefaultKeysFast(blocks);
   if (!blocks->Count && IsPGPError(err))
   {
      if (ForHourGlass) ForHourGlass->Cursor = crDefault;
      MsgBoxHwnd(owner->Handle,"Failed to import keys!","File Error",MB_OK);
      delete blocks;
      return false;
   }

   TPGPbuffer *key = new TPGPbuffer();
   TPGPbuffer *secrets = new TPGPbuffer();
   TPGPbuffer *publics = new TPGPbuffer();
   TPGPbuffer *newkeys = new TPGPbuffer(blocks);
   while (key->copykeyfrom(newkeys))
   {
         if (key->ispublic) publics->copyfrom(key);
         else secrets->copyfrom(key);
         key->clear();
   }
   publics->SaveToFile(PgpPublicRing());
   secrets->SaveToFile(PgpSecretRing());
   delete newkeys;
   delete secrets;
   delete publics;
   delete key;
   delete blocks;

   if (ForHourGlass) ForHourGlass->Cursor = crDefault;
   owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"");
   return true;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPsdk::TimeT2SigTime(time_t thetime)
{
   //Thu Oct 29 2:40:37 2001 (ctime)
   //10/29/01 2:40:37 PM     (sigtime)
   AnsiString mt,dy,yr,hr,mn,ampm="AM",timestr=ctime(&thetime);
   mt = s(s("~~~JanFebMarAprMayJunJulAugSepOctNovDec").Pos(GetToken(1,timestr," "))/3);
   dy = timestr.SubString(9,2);
   yr = timestr.SubString(23,2);
   mn = timestr.SubString(15,5);
   hr = timestr.SubString(12,2);
   while (hr.Length() && hr[1]=='0') hr = hr.SubString(2,2);
   int h = hr.ToIntDef(-1);
   if (h>12) { h -= 12; ampm = "PM"; }

   return mt+"/"+dy+"/"+yr+" "+s(h)+":"+mn+" "+ampm;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MessageDecryptVerifyLine(bool first, int decs, int sigs)
{
   AnsiString line = "*** "+s(first?"BEGIN":"END")+" PGP ";
   if (decs) line += "DECRYPTED";
   if (decs&&sigs) line += "/";
   if (sigs) line += "VERIFIED";
   line += " MESSAGE ***";
   return line;
}
//---------------------------------------------------------------------------
void __fastcall TPGPsdk::GrabHeaders(TStrings *message)
{
   headers->Clear();
   while (message->Count)
   {
      AnsiString line = message->Strings[0];
      headers->Add(line);
      message->Delete(0);
      if (line.IsEmpty())
         break;
   }
}
//---------------------------------------------------------------------------
void __fastcall TPGPsdk::PushHeaders(TStrings *message)
{
   for (int i=0; i<headers->Count; i++)
      message->Insert(i,headers->Strings[i]);
}
//---------------------------------------------------------------------------

PGPError TPGPsdk::DecodeEventHandler(PGPContextRef context, PGPEvent *event,
   PGPUserValue userValue)
{
   PGPError Err;
   Tdecinfo *decinfo = (Tdecinfo*)userValue;
   PGPEventSignatureData *sig=&event->data.signatureData;

   switch(event->type)
   {
      case kPGPEvent_DecryptionEvent:
         decinfo->decrypts++;
         break;
      case kPGPEvent_SignatureEvent:
      {
         char keyidstr[kPGPMaxKeyIDStringSize];
         Err=PGPsdk->PGPGetKeyIDString(&sig->signingKeyID,kPGPKeyIDString_Abbreviated,keyidstr);
         if (PGPsdk->IsPgpError(Err,"Failed to get key id string!"))
            return Err;

         AnsiString status,signer,sigtime;
         int idx = PGPsdk->pubkeys->IndexOf(keyidstr);
         if (decinfo->siginfos->Count) decinfo->siginfos->Add("");

         if (idx>=0) signer = GetToken(1,PGPsdk->pubkeys->Strings[idx],",");
         else signer = "Unknown, Key ID = "+s(keyidstr);

         sigtime = PGPsdk->TimeT2SigTime(PGPsdk->PGPGetStdTimeFromPGPTime(sig->creationTime));
         //if (sig->verified && sig->keyMeetsValidityThreshold) status = "good"; else

         if (sig->verified) status = "good"; else
         if (signer.Pos("Unknown")==1) status = "unknown"; else
            status = "bad";

         decinfo->siginfos->Add("*** PGP Signature Status: "+status);
         decinfo->siginfos->Add("*** Signer: "+signer);
         decinfo->siginfos->Add("*** Signed: "+sigtime);
         decinfo->siginfos->Add("*** Verified: "+FormatDateTime("m/dd/yy h:nn:ss ampm",Now()));
         decinfo->sigs++;
         break;
      }
   }
   return kPGPError_NoErr;
}
//---------------------------------------------------------------------------
void __fastcall TPGPsdk::CreateTempPubringFile()
{
   TPGPbuffer *pubs = new TPGPbuffer(PgpPublicRing());
   TPGPbuffer *rems = new TPGPbuffer(PgpRemailRing());
   pubs->Seek(0,soFromEnd); pubs->copyfrom(rems);
   pubs->SaveToFile(PgpPubTmpRing());
   delete rems;
   delete pubs;
}
//---------------------------------------------------------------------------
void __fastcall TPGPsdk::DeleteTempPubringFile()
{
   TStringList *files = new TStringList();
   GetFilenames(files,"pubring.*.tmp");
   for (int i=0; i<files->Count;i++)
      DeleteFile(files->Strings[i].c_str());
   delete files;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::VerifyClearSigned(TStrings *Message)
{
   SetCurrentDirectory(curdir);
   int i,idx;
   if ((idx=Message->IndexOf(BeginPgpSignedMessage))<0) return PGP_OK;
   headers->Clear(); for (int i=0; i<idx; i++)
      headers->Add(Message->Strings[i]);

   void *data=NULL; PGPSize size;
   PGPKeySetRef PubRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   Tdecinfo decinfo = { 0 }; decinfo.siginfos = new TStringList();

   CreateTempPubringFile();

   err = PGPNewFileSpecFromFullPath(context, PgpPubTmpRing().c_str(), &PubKeyRing);
   if(IsPgpError(err,"Failed to get PGP public keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_Mutable, PubKeyRing, &PubRingRef);
   if (IsPgpError(err,"Failed to open PGP public keyring"))
      goto Exit;

   err = PGPDecode(context,
      PGPOKeySetRef(context, PubRingRef),
      PGPOEventHandler(context, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOLastOption(context),
      0,0,0,0
   );
   if (IsPgpError(err,"Failed to decrypt message."))
      goto Exit;

   Message->SetText((char*)data);
   if (decinfo.decrypts || decinfo.sigs)
   {
      Message->Insert(0,MessageDecryptVerifyLine(1,decinfo.decrypts,decinfo.sigs));
      Message->Insert(1,"");
      for (i=0; i<decinfo.siginfos->Count; i++)
         Message->Insert(i,decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0,decinfo.decrypts,decinfo.sigs));
   }
   PushHeaders(Message);

Exit:
   delete decinfo.siginfos;
   if (data) PGPFreeData(data);
   if (PGPFileSpecRefIsValid(PubKeyRing)) PGPFreeFileSpec(PubKeyRing);
   if (PGPKeySetRefIsValid(PubRingRef)) PGPFreeKeySet(PubRingRef);
   DeleteTempPubringFile();
   return err==PGP_OK;
}
//---------------------------------------------------------------------------
/*AnsiString __fastcall TPGPsdk::MessageDecIdText8(TStrings *message)
{
   TPgpMsg *pgpmessage = new TPgpMsg(message);
   AnsiString keyids = pgpmessage->IdText8();
   delete pgpmessage;
   return keyids;
}*
//---------------------------------------------------------------------------
/*int pgp_packettype(BUFFER *b, long *len)
{
  int ctb = buf_getc(b);
  switch (ctb >> 6)
  {
     case 2:
        // old packet type
        switch (ctb & 3)
        {
           case 0: *len = buf_getc(b); break;
           case 1: *len = buf_geti(b); break;
           case 2: *len = buf_getl(b); break;
           case 3: *len = b->length - b->ptr; break;
        }
        return (ctb >> 2) & 15;
     case 3:
        // new packet type
        *len = buf_getc(b);
        if (*len >= 192 && *len <= 223) *len = (*len - 192) * 256 + buf_getc(b) + 192; else
        if (*len == 255) *len = buf_getl(b); else
        if (*len > 223) return (-2); // partial body length
        return (ctb & 63);
  }
  return (-1);
}
//---------------------------------------------------------------------------
int pgp_getpacket(BUFFER *in, BUFFER *p)
//  returns <0 = error, >0 = packet type
{
  long len;
  int type = pgp_packettype(in, &len);
  if (type>0) buf_get(in, p, len);
  return (type);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPgpMsg::idtext8(BUFFER *pkt)
{
   byte c;
   AnsiString text8;
   BUFFER *id = buf_new();
   KeyId(id,pkt);
    for (int i=0; i<4; i++)
   {
   	  c = id->data[4+i];
	  text8 += "0123456789ABCDEF" [ c >> 4  ];
	  text8 += "0123456789ABCDEF" [ c & 0xf ];
   }
   buf_free(id);
   return text8;
}*/
//---------------------------------------------------------------------------
//AnsiString __fastcall BufferDecIdText8(TMemoryStream *Buffer)
//{
   AnsiString keyids;
   /*int type;
   BUFFER *buf = buf_new();
   buf_append(buf,(char*)Buffer->Memory,(Buffer->Size>4096?4096:Buffer->Size));
   buf_rewind(buf);
   BUFFER *p = buf_new();
   while ((type = pgp_getpacket(buf, p)) > 0)
   {
      if (type==1)
      {
         if (!keyids.IsEmpty()) keyids += "\r\n";
         keyids += idtext8(p);
      }
   }*/
//   return keyids;
//}
//---------------------------------------------------------------------------
/*void __fastcall TPGPsdk::GetMessageStrLF(AnsiString &MsgStr, TStrings *Message)
{// obsolete function, unused
   MsgStr = "";
   for (int i=0; i<Message->Count; i++)
      MsgStr += Message->Strings[i]+"\n";
}*/
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::DecryptMessage(TStrings *Message, char *Passphrase,
   TStrings *DecKeyIds, bool recurse)
{
   SetCurrentDirectory(curdir);
   int i,idx;
   PGPKeyID KeyID;
   AnsiString keyidtext8;
   void *data=NULL; PGPSize size;
   PGPKeyRef DecKey = kInvalidPGPKeyRef;
   PGPKeySetRef KeyRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   Tdecinfo decinfo = { 0 }; decinfo.siginfos = new TStringList();
   TStringList *keyids = new TStringList();
   keyids->Text = DecKeyIds->Text; //MessageDecIdText8(Message);
   CreateTempPubringFile();

   err = PGPNewFileSpecFromFullPath(context, PgpPubTmpRing().c_str(), &PubKeyRing);
   if(IsPgpError(err,"Failed to get QS PGP public keyring file spec."))
      goto Exit;

   err = PGPNewFileSpecFromFullPath(context, PgpSecretRing().c_str(), &SecKeyRing);
   if(IsPgpError(err,"Failed to get QS PGP secret keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRingPair(context, false, PubKeyRing, SecKeyRing, &KeyRingRef);
   if (IsPgpError(err,"Failed to open QS PGP keyrings"))
      goto Exit;

   for (i=0; i<keyids->Count; i++)
   {
      keyidtext8 = keyids->Strings[i];
      err = PGPGetKeyIDFromString(keyidtext8.c_str(),&KeyID);
      if (IsntPGPError(err)) break;
   }
   if (IsPgpError(err,"Unable to get the required decryption keyid."))
      goto Exit;

   err = PGPGetKeyByKeyID(KeyRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&DecKey);
   if (IsPgpError(err,"Unable to get the required decryption key."))
      goto Exit;

   if (!PGPPassphraseIsValid(DecKey,PGPOPassphrase(context,Passphrase),PGPOLastOption(context)))
   {
      MessageBox(owner->Handle,
         "Sorry, You've entered a bad passphrase!","PGP Decrypt Error",MB_OK);
      err = kPGPError_BadPassphrase;
      goto Exit;
   }

   err = PGPDecode(context,
      PGPOKeySetRef(context, KeyRingRef),
      PGPOPassphrase(context, Passphrase),
      PGPORecursivelyDecode(context, recurse),
      PGPOEventHandler(context, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOLastOption(context),
      0,0
   );
   if (IsPgpError(err,"Failed to decrypt message."))
      goto Exit;

   Message->SetText((char*)data);
   if (decinfo.decrypts || decinfo.sigs)
   {
      // remove clearsig that sdk leaves for some reason
      if (decinfo.sigs && (idx=Message->IndexOf(BeginPgpSignedMessage))>=0)
      {
         // drop -----Begin Pgp... line and blank line next
         Message->Delete(idx);
         if (idx<Message->Count) Message->Delete(idx);
         // find -----Begin Pgp Sig...
         if ((idx=Message->IndexOf(BeginPgpSignature))>=0)
         {
            // drop blank before and rest of msg
            if (idx>0) idx--; while (idx<Message->Count)
               Message->Delete(idx);
         }
      }
      Message->Insert(0,MessageDecryptVerifyLine(1,decinfo.decrypts,decinfo.sigs));
      Message->Insert(1,"");
      for (i=0; i<decinfo.siginfos->Count; i++)
         Message->Insert(i,decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0,decinfo.decrypts,decinfo.sigs));
   }

Exit:
   delete decinfo.siginfos;
   if (data) PGPFreeData(data);
   if (PGPFileSpecRefIsValid(SecKeyRing)) PGPFreeFileSpec(SecKeyRing);
   if (PGPFileSpecRefIsValid(PubKeyRing)) PGPFreeFileSpec(PubKeyRing);
   if (PGPKeySetRefIsValid(KeyRingRef)) PGPFreeKeySet(KeyRingRef);
   DeleteTempPubringFile();
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------
/*PGPError __fastcall TPGPsdk::DecryptBuffer(TMemoryStream *Buffer, char *Passphrase,
   bool recurse)
{
   int i;//,idx;
   PGPKeyID KeyID;
   AnsiString keyidtext8;
   void *data=NULL; PGPSize size;
   PGPKeyRef DecKey = kInvalidPGPKeyRef;
   PGPKeySetRef KeyRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   Tdecinfo decinfo = { 0 }; decinfo.siginfos = new TStringList();
   TStringList *keyids = new TStringList();
   keyids->Text = BufferDecIdText8(Buffer);
   CreateTempPubringFile();

   err = PGPNewFileSpecFromFullPath(context, PgpPubTmpRing().c_str(), &PubKeyRing);
   if(IsPgpError(err,"Failed to get QS PGP public keyring file spec."))
      goto Exit;

   err = PGPNewFileSpecFromFullPath(context, PgpSecretRing().c_str(), &SecKeyRing);
   if(IsPgpError(err,"Failed to get QS PGP secret keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRingPair(context, false, PubKeyRing, SecKeyRing, &KeyRingRef);
   if (IsPgpError(err,"Failed to open QS PGP keyrings"))
      goto Exit;

   for (i=0; i<keyids->Count; i++)
   {
      keyidtext8 = keyids->Strings[i];
      err = PGPGetKeyIDFromString(keyidtext8.c_str(),&KeyID);
      if (IsntPGPError(err)) break;
   }
   if (IsPgpError(err,"Unable to get the required decryption keyid."))
      goto Exit;

   err = PGPGetKeyByKeyID(KeyRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&DecKey);
   if (IsPgpError(err,"Unable to get the required decryption key."))
      goto Exit;

   if (!PGPPassphraseIsValid(DecKey,PGPOPassphrase(context,Passphrase),PGPOLastOption(context)))
   {
      MessageBox(owner->Handle,
         "Sorry, You've entered a bad passphrase!","PGP Decrypt Error",MB_OK);
      err = kPGPError_BadPassphrase;
      goto Exit;
   }

   err = PGPDecode(context,
      PGPOKeySetRef(context, KeyRingRef),
      PGPOPassphrase(context, Passphrase),
      PGPORecursivelyDecode(context, recurse),
      PGPOEventHandler(context, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(context, Buffer->Memory, Buffer->Size),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOLastOption(context),
      0,0
   );
   if (IsPgpError(err,"Failed to decrypt message."))
      goto Exit;

   Buffer->SetSize(0);
   Buffer->WriteBuffer(data,size);

Exit:
   delete decinfo.siginfos;
   if (data) PGPFreeData(data);
   if (PGPFileSpecRefIsValid(SecKeyRing)) PGPFreeFileSpec(SecKeyRing);
   if (PGPFileSpecRefIsValid(PubKeyRing)) PGPFreeFileSpec(PubKeyRing);
   if (PGPKeySetRefIsValid(KeyRingRef)) PGPFreeKeySet(KeyRingRef);
   DeleteTempPubringFile();
   return err;
}*/
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::SignMessage(TStrings *Message, char *SignKeyid,
   char *Passphrase)
{
   SetCurrentDirectory(curdir);
   PGPKeyID KeyID;
   AnsiString line,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef SignKey = kInvalidPGPKeyRef;
   PGPKeySetRef SecRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   PGPHashAlgorithm HashAlg;

   GrabHeaders(Message);
   err = PGPNewFileSpecFromFullPath(context, PgpSecretRing().c_str(), &SecKeyRing);
   if(IsPgpError(err,"Failed to get PGP secret keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_Mutable, SecKeyRing, &SecRingRef);
   if (IsPgpError(err,"Failed to open PGP secret keyring"))
      goto Exit;

   err = PGPGetKeyIDFromString(SignKeyid,&KeyID);
   if (IsPgpError(err,"Unable to get the required signing keyid."))
      goto Exit;

   err = PGPGetKeyByKeyID(SecRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&SignKey);
   if (IsPgpError(err,"Unable to get the required signing key."))
      goto Exit;

   err = PGPGetHashAlgUsed(SignKey,&HashAlg);
   if (IsPgpError(err,"Unable to get the signing key hash algorithm."))
      goto Exit;

   if (HashAlg==kPGPHashAlgorithm_Invalid) HashAlg = kPGPHashAlgorithm_SHA;

   /*SYSTEMTIME prevtime,oddtime;
   if (SIGNWITHINVALIDDATE)
   {
   time_t t = PGPGetStdTimeFromPGPTime(PGPGetTime());
   sm(ctime(&t));
      GetLocalTime(&oddtime); GetSystemTime(&prevtime);
   sm(s((int)oddtime.wHour)+"   "+s((int)oddtime.wMinute)+"   "+s((int)oddtime.wSecond));
   sm(s((int)prevtime.wHour)+"   "+s((int)prevtime.wMinute)+"   "+s((int)prevtime.wSecond));
      //oddtime.wHour = 0;
      //oddtime.wMinute = 0;
      //oddtime.wSecond = 0;
      //oddtime.wMilliseconds = 1;
      //if (!SetSystemTime(&oddtime)) sm(GetLastErrorStr());
      //sm(s((int)oddtime.wHour)+"   "+s((int)oddtime.wMinute)+"   "+s((int)oddtime.wSecond));
   }*/
   err = PGPEncode(context,
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOSignWithKey( context, SignKey,
         PGPOPassphrase( context, Passphrase ),
         PGPOLastOption( context ),0),
      PGPOHashAlgorithm( context, HashAlg),
      PGPOClearSign( context, true ),
      PGPOVersionString( context, "N/A" ),
      PGPODataIsASCII( context, true ),
      PGPOLastOption( context ),
      0,0,0,0,0,0,0,0
   );
   /*if (SIGNWITHINVALIDDATE)
   {
      SetSystemTime(&prevtime);
   }*/
   if (IsPgpError(err,"Failed to clearsign message."))
      goto Exit;

   Message->SetText((char*)data);
   PushHeaders(Message);

Exit:
   if (PGPFileSpecRefIsValid(SecKeyRing)) PGPFreeFileSpec(SecKeyRing);
   if (PGPKeySetRefIsValid(SecRingRef)) PGPFreeKeySet(SecRingRef);
   if (data) PGPFreeData(data);
   return err;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::EncryptSignMessage(TStrings *Message, TStrings *EncKeyIds,
   char *SignKeyid, char *Passphrase)
{
   SetCurrentDirectory(curdir);
   int i;
   PGPKeyID KeyID;
   AnsiString line,keyid,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef EncKey = kInvalidPGPKeyRef;
   PGPKeyRef SignKey = kInvalidPGPKeyRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   PGPKeySetRef TheRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef EncRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef SinRingRef = kInvalidPGPKeySetRef;
   PGPHashAlgorithm HashAlg;

   GrabHeaders(Message);
   CreateTempPubringFile();

   err = PGPNewFileSpecFromFullPath(context, PgpPubTmpRing().c_str(), &PubKeyRing);
   if(IsPgpError(err,"Failed to get PGP public keyring file spec."))
      goto Exit;
   err = PGPNewFileSpecFromFullPath(context, PgpSecretRing().c_str(), &SecKeyRing);
   if(IsPgpError(err,"Failed to get PGP secret keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRingPair(context, kPGPKeyRingOpenFlags_Mutable, PubKeyRing,
      SecKeyRing, &TheRingRef);

   if (IsPgpError(err,"Failed to open pgp keyrings"))
      goto Exit;

   err = PGPNewKeySet(context,&EncRingRef);
   if (IsPgpError(err,"Failed creating keyset of requested keyids"))
      goto Exit;

   i=-1; while(++i<EncKeyIds->Count)
   {
      keyid = EncKeyIds->Strings[i].SubString(3,8);
      err = PGPGetKeyIDFromString(keyid.c_str(),&KeyID);
      if (IsPgpError(err,"Unable to get the required encrypting keyid."))
         goto Exit;

      err = PGPGetKeyByKeyID(TheRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&EncKey);
      if (IsPgpError(err,"Unable to get the required encrypting key."))
         goto Exit;

      err = PGPNewSingletonKeySet(EncKey,&SinRingRef);
      if (IsPgpError(err,s("Failed creating singlton keyset for key "+keyid).c_str()))
         goto Exit;

      PGPAddKeys(SinRingRef,EncRingRef);
      PGPFreeKeySet(SinRingRef);
      SinRingRef = kInvalidPGPKeySetRef;
   }

   err = PGPGetKeyIDFromString(SignKeyid,&KeyID);
   if (IsPgpError(err,"Unable to get the required signing keyid."))
      goto Exit;

   err = PGPGetKeyByKeyID(TheRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&SignKey);
   if (IsPgpError(err,"Unable to get the required signing key."))
      goto Exit;

   err = PGPGetHashAlgUsed(SignKey,&HashAlg);
   if (IsPgpError(err,"Unable to get the signing key hash algorithm."))
      goto Exit;

   err = PGPEncode( context,
      PGPOEncryptToKeySet( context, EncRingRef ),
      PGPOSignWithKey( context, SignKey,
         PGPOPassphrase( context, Passphrase ),
         PGPOLastOption( context ),0),
      PGPOHashAlgorithm( context, HashAlg),
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0
   );

   if (IsPgpError(err,"Failed to encrypt message"))
      goto Exit;

   Message->SetText((char*)data);
   PushHeaders(Message);

Exit:
   if (PGPFileSpecRefIsValid(PubKeyRing)) PGPFreeFileSpec(PubKeyRing);
   if (PGPFileSpecRefIsValid(SecKeyRing)) PGPFreeFileSpec(SecKeyRing);
   if (PGPKeySetRefIsValid(TheRingRef)) PGPFreeKeySet(TheRingRef);
   if (PGPKeySetRefIsValid(EncRingRef)) PGPFreeKeySet(EncRingRef);
   if (PGPKeySetRefIsValid(SinRingRef)) PGPFreeKeySet(SinRingRef);
   DeleteTempPubringFile();
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::CEncryptMessage(TStrings *Message, char *Passphrase)
{
   SetCurrentDirectory(curdir);
   AnsiString line,msgstr;
   void *data=NULL; PGPSize size;

   GrabHeaders(Message);
   err = PGPEncode( context,
      PGPOConventionalEncrypt(context,
         PGPOPassphrase( context, Passphrase ),
         PGPOLastOption( context ),
         0),
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0,0,0
   );
   if (IsPgpError(err,"Failed to conventionally encrypt message"))
      goto Exit;

   Message->SetText((char*)data);
   PushHeaders(Message);

Exit:
   if (data) PGPFreeData(data);
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------
bool __fastcall TPGPsdk::EncryptMessage(TStrings *Message, TStrings *KeyIds)
{
   SetCurrentDirectory(curdir);
   int i;
   PGPKeyID KeyID;
   AnsiString line,keyid,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef EncKey = kInvalidPGPKeyRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPKeySetRef PubRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef EncRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef SinRingRef = kInvalidPGPKeySetRef;

   GrabHeaders(Message);
   CreateTempPubringFile();

   err = PGPNewFileSpecFromFullPath(context, PgpPubTmpRing().c_str(), &PubKeyRing);
   if(IsPgpError(err,"Failed to get PGP public keyring file spec."))
      goto Exit;

   err = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_Mutable, PubKeyRing, &PubRingRef);
   if (IsPgpError(err,"Failed to open PGP public keyring"))
      goto Exit;

   err = PGPNewKeySet(context,&EncRingRef);
   if (IsPgpError(err,"Failed creating keyset of requested keyids"))
      goto Exit;

   i=-1; while(++i<KeyIds->Count)
   {
      keyid = KeyIds->Strings[i].SubString(3,8);
      err = PGPGetKeyIDFromString(keyid.c_str(),&KeyID);
      if (IsPgpError(err,"Unable to get the required encrypting keyid."))
         goto Exit;

      err = PGPGetKeyByKeyID(PubRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&EncKey);
      if (IsPgpError(err,"Unable to get the required encrypting key."))
         goto Exit;

      err = PGPNewSingletonKeySet(EncKey,&SinRingRef);
      if (IsPgpError(err,s("Failed creating singlton keyset for key "+keyid).c_str()))
         goto Exit;

      PGPAddKeys(SinRingRef,EncRingRef);
      PGPFreeKeySet(SinRingRef);
      SinRingRef = kInvalidPGPKeySetRef;
   }

   err = PGPEncode( context,
      PGPOEncryptToKeySet( context, EncRingRef ),
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0,0,0
   );
   if (IsPgpError(err,"Failed to encrypt message"))
      goto Exit;

   Message->SetText((char*)data);
   PushHeaders(Message);

Exit:
   if (PGPFileSpecRefIsValid(PubKeyRing)) PGPFreeFileSpec(PubKeyRing);
   if (PGPKeySetRefIsValid(PubRingRef)) PGPFreeKeySet(PubRingRef);
   if (PGPKeySetRefIsValid(EncRingRef)) PGPFreeKeySet(EncRingRef);
   if (PGPKeySetRefIsValid(SinRingRef)) PGPFreeKeySet(SinRingRef);
   DeleteTempPubringFile();
   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk::IsPgpError(PGPError err, char *errmsg)
{
   if (IsPGPError(err))
   {
      ShowPGPError(err, errmsg);
      return true;
   }
   return false;
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetErrorString(PGPError theError, PGPSize bufferSize,
   char *theString)
{
   PGPGETERRORSTRING geterrstr =
      (PGPGETERRORSTRING)GetProcAddress(hSdkLib, "PGPGetErrorString");
   assert(geterrstr != NULL);

   return geterrstr(theError, bufferSize, theString);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPsdkInit()
{
   PGPSDKINIT sdkinit = (PGPSDKINIT)GetProcAddress(hSdkLib,"PGPsdkInit");
   assert(sdkinit != NULL);

   return sdkinit();
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPsdkCleanup()
{
   PGPSDKCLEANUP sdkcleanup = (PGPSDKCLEANUP)GetProcAddress(hSdkLib, "PGPsdkCleanup");
   assert(sdkcleanup != NULL);

   return sdkcleanup();
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPNewContext(PGPUInt32 clientAPIVersion,
   PGPContextRef *newContext)
{
   PGPNEWCONTEXT newcontext =
      (PGPNEWCONTEXT)GetProcAddress(hSdkLib,"PGPNewContext");
   assert(newcontext != NULL);

   return newcontext(clientAPIVersion, newContext);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPsdkLoadDefaultPrefs(PGPContextRef context)
{
   PGPSDKLOADDEFAULTPREFS loadprefs =
      (PGPSDKLOADDEFAULTPREFS)GetProcAddress(hSdkLib,"PGPsdkLoadDefaultPrefs");
   assert(loadprefs != NULL);

   return loadprefs(context);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPsdkPrefGetFileSpec(PGPContextRef context,
   PGPsdkPrefSelector selector, PGPFileSpecRef * outRef)
{
   PGPSDKPREFGETFILESPEC filespec =
      (PGPSDKPREFGETFILESPEC)GetProcAddress(hSdkLib, "PGPsdkPrefGetFileSpec");
   assert(filespec != NULL);

   return filespec(context, selector, outRef);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPGetFullPathFromFileSpec(PGPFileSpecRef fileRef,
   char **fullPathPtr)
{
   PGPGETFULLPATHFROMFILESPEC fullpath =
     (PGPGETFULLPATHFROMFILESPEC)GetProcAddress(hSdkLib,"PGPGetFullPathFromFileSpec");
   assert(fullpath != NULL);

   return fullpath(fileRef, fullPathPtr);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPGetSDKVersion(PGPUInt32 *version)
{
   PGPGETSDKVERSION sdkversion =
      (PGPGETSDKVERSION)GetProcAddress(hSdkLib,"PGPGetSDKVersion");
   assert(sdkversion != NULL);

   return  sdkversion(version);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPFreeFileSpec(PGPFileSpecRef fileRef)
{
   PGPFREEFILESPEC freespec = 
      (PGPFREEFILESPEC)GetProcAddress(hSdkLib,"PGPFreeFileSpec");
   return freespec(fileRef);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPExportKeySet(PGPKeySetRef keys, PGPOptionListRef firstOpt,
   PGPOptionListRef secondOpt, PGPOptionListRef thirdOpt, PGPOptionListRef forthOpt,
   PGPOptionListRef fifthOpt)
{
   PGPEXPORTKEYSET exportset =
      (PGPEXPORTKEYSET)GetProcAddress(hSdkLib,"PGPExportKeySet");
   assert(exportset != NULL);

   return exportset(keys, firstOpt, secondOpt, thirdOpt, forthOpt, fifthOpt);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOExportFormat(PGPContextRef context,
   PGPExportFormat exportFormat)
{
   PGPOEXPORTFORMAT exportfmt =
     (PGPOEXPORTFORMAT)GetProcAddress(hSdkLib,"PGPOExportFormat");

   return exportfmt(context, exportFormat);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOExportPrivateKeys(PGPContextRef context,
   PGPBoolean exportKeys)
{
   PGPOEXPORTPRIVATEKEYS exportprivs =
      (PGPOEXPORTPRIVATEKEYS)GetProcAddress(hSdkLib,"PGPOExportPrivateKeys");
   assert (exportprivs != NULL);

   return exportprivs(context, exportKeys);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOVersionString(PGPContextRef context,
   char const *version)
{
   PGPOVERSIONSTRING verstr =
      (PGPOVERSIONSTRING)GetProcAddress(hSdkLib,"PGPOVersionString");
   assert(verstr != NULL);

   return verstr(context, version);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOLastOption(PGPContextRef context)
{
   PGPOLASTOPTION lastopt =
      (PGPOLASTOPTION)GetProcAddress(hSdkLib,"PGPOLastOption");
   assert(lastopt != NULL);

   return lastopt(context);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOAllocatedOutputBuffer(PGPContextRef context,
   void **buffer, PGPSize maximumBufferSize, PGPSize *actualBufferSize)
{
   PGPOALLOCATEDOUTPUTBUFFER allbuf =
      (PGPOALLOCATEDOUTPUTBUFFER)GetProcAddress(hSdkLib,"PGPOAllocatedOutputBuffer");
   assert(allbuf != NULL);

   return allbuf(context, buffer, maximumBufferSize, actualBufferSize);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPORecursivelyDecode(PGPContextRef context,
   PGPBoolean recurse)
{
   PGPORECURSIVELYDECODE recur =
      (PGPORECURSIVELYDECODE)GetProcAddress(hSdkLib,"PGPORecursivelyDecode");
   assert(recur != NULL);

   return recur(context,recurse);
}
//---------------------------------------------------------------------------
PGPOptionListRef __fastcall TPGPsdk::PGPOPassphrase(PGPContextRef context,
   const char *passphrase)
{
   PGPOPASSPHRASE phrase = (PGPOPASSPHRASE)GetProcAddress(hSdkLib,"PGPOPassphrase");
   assert(phrase != NULL);

   return phrase(context,passphrase);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPFreeKeySet(PGPKeySetRef keys)
{
   PGPFREEKEYSET freeset = (PGPFREEKEYSET)GetProcAddress(hSdkLib,"PGPFreeKeySet");
   assert(freeset != NULL);

   return freeset(keys);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPFreeData(void *allocation)
{
   PGPFREEDATA freedat = (PGPFREEDATA)GetProcAddress(hSdkLib,"PGPFreeData");
   assert(freedat != NULL);

   return freedat(allocation);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPOpenDefaultKeyRings(PGPContextRef context,
   PGPKeyRingOpenFlags openFlags, PGPKeySetRef *keySet)
{
   PGPOPENDEFAULTKEYRINGS opendefs =
      (PGPOPENDEFAULTKEYRINGS)GetProcAddress(hSdkLib,"PGPOpenDefaultKeyRings");
   assert(opendefs != NULL);

   return opendefs(context, openFlags, keySet);
}
//---------------------------------------------------------------------------
PGPError __fastcall TPGPsdk::PGPNewKeyDisabledFilter(PGPContextRef context,
   PGPBoolean disabled, PGPFilterRef *outFilter)
{
   PGPNEWKEYDISABLEDFILTER filter =
      (PGPNEWKEYDISABLEDFILTER)GetProcAddress(hSdkLib,"PGPNewKeyDisabledFilter");
   assert(filter != NULL);

   return filter(context, disabled, outFilter);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPNewKeyRevokedFilter(PGPContextRef context, PGPBoolean revoked,
      PGPFilterRef *outFilter)
{
   PGPNEWKEYREVOKEDFILTER filter =
      (PGPNEWKEYREVOKEDFILTER)GetProcAddress(hSdkLib,"PGPNewKeyRevokedFilter");
   assert(filter != NULL);

   return filter(context, revoked, outFilter);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPFilterKeySet(PGPKeySetRef origSet, PGPFilterRef filter,
      PGPKeySetRef *resultSet)
{
   PGPFILTERKEYSET filterset =
      (PGPFILTERKEYSET)GetProcAddress(hSdkLib,"PGPFilterKeySet");
   assert(filterset != NULL);

   return filterset(origSet, filter, resultSet);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPOrderKeySet(PGPKeySetRef src, PGPKeyOrdering order,
      PGPKeyListRef *outRef )
{
   PGPORDERKEYSET orderset =
      (PGPORDERKEYSET)GetProcAddress(hSdkLib,"PGPOrderKeySet");
   assert(orderset != NULL);

   return orderset(src, order, outRef);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPNewKeyIter(PGPKeyListRef keys, PGPKeyIterRef *outRef)
{
   PGPNEWKEYITER iter =
      (PGPNEWKEYITER)GetProcAddress(hSdkLib,"PGPNewKeyIter");
   assert(iter != NULL);

   return iter(keys, outRef);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetKeyIDFromKey(PGPKeyRef key, PGPKeyID *id)
{
   PGPGETKEYIDFROMKEY getid =
      (PGPGETKEYIDFROMKEY)GetProcAddress(hSdkLib,"PGPGetKeyIDFromKey");
   return getid(key, id);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetHashAlgUsed(PGPKeyRef key, PGPHashAlgorithm *hashAlg )
{
   PGPGETHASHALGUSED gethash =
      (PGPGETHASHALGUSED)GetProcAddress(hSdkLib,"PGPGetHashAlgUsed");
   assert(gethash != NULL);

   return gethash(key, hashAlg);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetKeyIDFromString(const char *string, PGPKeyID *id)
{
   PGPGETKEYIDFROMSTRING getid =
      (PGPGETKEYIDFROMSTRING)GetProcAddress(hSdkLib,"PGPGetKeyIDFromString");
   assert(getid != NULL);

   return getid(string, id);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetKeyIDString(PGPKeyID const *ref, PGPKeyIDStringType type,
      char outString[kPGPMaxKeyIDStringSize])
{
   PGPGETKEYIDSTRING getstr =
      (PGPGETKEYIDSTRING)GetProcAddress(hSdkLib,"PGPGetKeyIDString");
   assert(getstr != NULL);

   return getstr(ref, type, outString);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPKeyIterNext(PGPKeyIterRef iter, PGPKeyRef *outRef)
{
   PGPKEYITERNEXT iternxt = (PGPKEYITERNEXT)GetProcAddress(hSdkLib,"PGPKeyIterNext");
   assert(iternxt != NULL);

   return iternxt(iter, outRef);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPFreeKeyIter(PGPKeyIterRef iter)
{
   PGPFREEKEYITER freeiter =
      (PGPFREEKEYITER)GetProcAddress(hSdkLib,"PGPFreeKeyIter");
   assert(freeiter != NULL);

   return freeiter(iter);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPFreeKeyList(PGPKeyListRef keys)
{
   PGPFREEKEYLIST freelist = (PGPFREEKEYLIST)GetProcAddress(hSdkLib,"PGPFreeKeyList");
   assert(freelist != NULL);

   return freelist(keys);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPFreeFilter(PGPFilterRef filter)
{
   PGPFREEFILTER freefilter = (PGPFREEFILTER)GetProcAddress(hSdkLib,"PGPFreeFilter");
   assert(freefilter != NULL);

   return freefilter(filter);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPNewFileSpecFromFullPath(PGPContextRef context,
   const char *path, PGPFileSpecRef *ref)
{
   PGPNEWFILESPECFROMFULLPATH getspec =
      (PGPNEWFILESPECFROMFULLPATH)GetProcAddress(hSdkLib,"PGPNewFileSpecFromFullPath");
   assert(getspec != NULL);

   return getspec(context,path,ref);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPOpenKeyRing(PGPContextRef context, PGPKeyRingOpenFlags openFlags,
      PGPFileSpecRef fileRef, PGPKeySetRef *keySet)
{
   PGPOPENKEYRING openring = (PGPOPENKEYRING)GetProcAddress(hSdkLib,"PGPOpenKeyRing");
   assert(openring != NULL);

   return openring(context,openFlags,fileRef,keySet);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOInputBuffer( PGPContextRef context,
   void const *buffer, PGPSize bufferSize)
{
   PGPOINPUTBUFFER inbuf = (PGPOINPUTBUFFER)GetProcAddress(hSdkLib,"PGPOInputBuffer");
   assert(inbuf != NULL);

   return inbuf(context, buffer, bufferSize);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOKeySetRef(PGPContextRef context, PGPKeySetRef keysetRef)
{
   PGPOKEYSETREF setref = (PGPOKEYSETREF)GetProcAddress(hSdkLib,"PGPOKeySetRef");
   assert(setref != NULL);

   return setref(context, keysetRef);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPOpenKeyRingPair(PGPContextRef context, PGPKeyRingOpenFlags openFlags,
   PGPFileSpecRef pubRef, PGPFileSpecRef secRef, PGPKeySetRef *keySet)
{
   PGPOPENKEYRINGPAIR openpair =
      (PGPOPENKEYRINGPAIR)GetProcAddress(hSdkLib,"PGPOpenKeyRingPair");
   assert(openpair != NULL);

   return openpair(context,openFlags,pubRef,secRef,keySet);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetKeyByKeyID(PGPKeySetRef keySet, PGPKeyID const *keyID,
      PGPPublicKeyAlgorithm pubKeyAlgorithm, PGPKeyRef *key )
{
   PGPGETKEYBYKEYID getkey =
      (PGPGETKEYBYKEYID)GetProcAddress(hSdkLib,"PGPGetKeyByKeyID");
   assert(getkey != NULL);

   return getkey(keySet, keyID, pubKeyAlgorithm, key);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPDecodeSig(PGPContextRef context, PGPOptionListRef firstOption,
   PGPOptionListRef secondOption, PGPOptionListRef thirdOption, PGPOptionListRef forthOption,
   PGPOptionListRef fifthOption)
{
   PGPDECODESIG decode = (PGPDECODESIG)GetProcAddress(hSdkLib,"PGPDecode");
   assert(decode != NULL);

   return decode(context, firstOption, secondOption, thirdOption, forthOption,
         fifthOption);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPDecode(PGPContextRef context, PGPOptionListRef opt1,
   PGPOptionListRef opt2, PGPOptionListRef opt3, PGPOptionListRef opt4,
   PGPOptionListRef opt5, PGPOptionListRef opt6, PGPOptionListRef opt7,
   PGPOptionListRef opt8, PGPOptionListRef opt9)
{
   PGPDECODE decode = (PGPDECODE)GetProcAddress(hSdkLib,"PGPDecode");
   assert(decode != NULL);

   return decode(context, opt1, opt2, opt3, opt4, opt5, opt6, opt7, opt8, opt9);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPEncode(PGPContextRef context, PGPOptionListRef opt1,
   PGPOptionListRef opt2, PGPOptionListRef opt3, PGPOptionListRef opt4,
   PGPOptionListRef opt5, PGPOptionListRef opt6, PGPOptionListRef opt7,
   PGPOptionListRef opt8, PGPOptionListRef opt9, PGPOptionListRef opt10,
   PGPOptionListRef opt11, PGPOptionListRef opt12, PGPOptionListRef opt13,
   PGPOptionListRef opt14, PGPOptionListRef opt15,
   PGPOptionListRef opt16)
{
   PGPENCODE encode = (PGPENCODE)GetProcAddress(hSdkLib,"PGPEncode");
   assert(encode != NULL);

   return encode(context, opt1, opt2, opt3, opt4, opt5, opt6, opt7, opt8,
        opt9, opt10, opt11, opt12, opt13, opt14, opt15, opt16);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOCompression(PGPContextRef pgpContext,
      PGPBoolean isCompressed )
{
   PGPOCOMPRESSION comp = (PGPOCOMPRESSION)GetProcAddress(hSdkLib,"PGPOCompression");
   assert(comp != NULL);

   return comp(context, isCompressed);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOClearSign(PGPContextRef context, PGPBoolean clearsign)
{
   PGPOCLEARSIGN sign = (PGPOCLEARSIGN)GetProcAddress(hSdkLib,"PGPOClearSign");
   assert(sign != NULL);

   return sign(context, clearsign);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOSignWithKey(PGPContextRef context, PGPKeyRef SignKey,
   PGPOptionListRef opt1, PGPOptionListRef opt2, PGPOptionListRef opt3)
{
   PGPOSIGNWITHKEY sign = (PGPOSIGNWITHKEY)GetProcAddress(hSdkLib,"PGPOSignWithKey");
   assert(sign != NULL);

   return sign(context, SignKey, opt1, opt2, opt3);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPNewKeySet(PGPContextRef context, PGPKeySetRef *keySet)
{
   PGPNEWKEYSET newset = (PGPNEWKEYSET)GetProcAddress(hSdkLib,"PGPNewKeySet");
   assert(newset != NULL);

   return newset(context, keySet);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPNewSingletonKeySet(PGPKeyRef key, PGPKeySetRef *keySet)
{
   PGPNEWSINGLETONKEYSET newset =
      (PGPNEWSINGLETONKEYSET)GetProcAddress(hSdkLib,"PGPNewSingletonKeySet");
   assert(newset != NULL);

   return newset(key, keySet);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPAddKeys(PGPKeySetRef keysToAdd, PGPKeySetRef keySet)
{
   PGPADDKEYS add = (PGPADDKEYS)GetProcAddress(hSdkLib,"PGPAddKeys");
   assert(add != NULL);

   return add(keysToAdd, keySet);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOArmorOutput(PGPContextRef Context,
   PGPBoolean armorOutput)
{
   PGPOARMOROUTPUT armor =
      (PGPOARMOROUTPUT)GetProcAddress(hSdkLib,"PGPOArmorOutput");
   assert(armor != NULL);

   return armor(Context,armorOutput);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOEncryptToKeySet(PGPContextRef Context,
   PGPKeySetRef keySet )
{
   PGPOENCRYPTTOKEYSET encset =
      (PGPOENCRYPTTOKEYSET)GetProcAddress(hSdkLib,"PGPOEncryptToKeySet");
   assert(encset != NULL);

   return encset(Context,keySet);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOConventionalEncrypt(PGPContextRef Context,
      PGPOptionListRef opt1, PGPOptionListRef opt2, PGPOptionListRef opt3)
{
   PGPOCONVENTIONALENCRYPT enccon =
      (PGPOCONVENTIONALENCRYPT)GetProcAddress(hSdkLib,"PGPOConventionalEncrypt");
   assert(enccon != NULL);

   return enccon(Context,opt1,opt2,opt3);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOOutputLineEndType(PGPContextRef context,
   PGPLineEndType lineEndType )
{
   PGPOOUTPUTLINEENDTYPE lend =
      (PGPOOUTPUTLINEENDTYPE)GetProcAddress(hSdkLib,"PGPOOutputLineEndType");
   assert(lend != NULL);

   return lend(context,lineEndType);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOForYourEyesOnly(PGPContextRef pgpContext,
   PGPBoolean forYourEyesOnly)
{
   PGPOFORYOUREYESONLY eyesonly =
      (PGPOFORYOUREYESONLY)GetProcAddress(hSdkLib,"PGPOForYourEyesOnly");
   assert(eyesonly != NULL);

   return eyesonly(pgpContext,forYourEyesOnly);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOHashAlgorithm(PGPContextRef pgpContext,
   PGPHashAlgorithm algID )
{
   PGPOHASHALGORITHM hashalg =
      (PGPOHASHALGORITHM)GetProcAddress(hSdkLib,"PGPOHashAlgorithm");
   assert(hashalg != NULL);

   return hashalg(pgpContext,algID);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPODataIsASCII(PGPContextRef pgpContext,
   PGPBoolean dataIsASCII )
{
   PGPODDATAISSCII ascii = (PGPODDATAISSCII)GetProcAddress(hSdkLib,"PGPODataIsASCII");
   assert(ascii != NULL);

   return ascii(context,dataIsASCII);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPCommitKeyRingChanges( PGPKeySetRef keySet )
{
   PGPCOMMITKEYRINGCHANGES commit =
      (PGPCOMMITKEYRINGCHANGES)GetProcAddress(hSdkLib,"PGPCommitKeyringChanges");
   assert(commit != NULL);

   return commit(keySet);
}
//---------------------------------------------------------------------------

PGPOptionListRef __fastcall TPGPsdk::PGPOEventHandler(PGPContextRef context,
       PGPEventHandlerProcPtr eventHandler, PGPUserValue eventHandlerData)
{
   PGPOEVENTHANDLER handler =
      (PGPOEVENTHANDLER)GetProcAddress(hSdkLib,"PGPOEventHandler");
   assert(handler != NULL);

   return handler(context, eventHandler, eventHandlerData);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetPrimaryUserID(PGPKeyRef key, PGPUserIDRef *userID)
{
   PGPGETPRIMARYUSERID getid =
      (PGPGETPRIMARYUSERID)GetProcAddress(hSdkLib,"PGPGetPrimaryUserID");
   assert(getid != NULL);

   return getid(key, userID);
}
//---------------------------------------------------------------------------

PGPError __fastcall TPGPsdk::PGPGetUserIDStringBuffer(PGPUserIDRef userID,
   PGPUserIDPropName propName, PGPSize availLength, char *propString,
   PGPSize *usedLength)
{
   PGPGETUSERIDSTRINGBUFFER getstr =
      (PGPGETUSERIDSTRINGBUFFER)GetProcAddress(hSdkLib,"PGPGetUserIDStringBuffer");
   assert(getstr != NULL);

   return getstr(userID, propName, availLength, propString, usedLength);
}
//---------------------------------------------------------------------------

PGPBoolean __fastcall TPGPsdk::PGPPassphraseIsValid(PGPKeyRef key, PGPOptionListRef opt1,
    PGPOptionListRef opt2)
{
   PGPPASSPHRASEISVALID goodpass =
      (PGPPASSPHRASEISVALID)GetProcAddress(hSdkLib,"PGPPassphraseIsValid");
   assert(goodpass != NULL);

   return goodpass(key,opt1,opt2);
}
//---------------------------------------------------------------------------

time_t __fastcall TPGPsdk::PGPGetStdTimeFromPGPTime(PGPTime theTime)
{
   PGPGETSTDTIMEFROMPGPTIME convtime =
      (PGPGETSTDTIMEFROMPGPTIME)GetProcAddress(hSdkLib,"PGPGetStdTimeFromPGPTime");
   assert(convtime != NULL);

   return convtime(theTime);
}
//---------------------------------------------------------------------------

PGPTime __fastcall TPGPsdk::PGPGetTime()
{
   PGPGETTIME gettime = (PGPGETTIME)GetProcAddress(hSdkLib,"PGPGetTime");
   assert(gettime != NULL);

   return gettime();
}
//---------------------------------------------------------------------------





